"""
Refiner answer deserialization into boiler models.
"""
from itertools import groupby
import json
from typing import Any, Dict, List, Union

from intervals import IntInterval

from boiler import BoilerError
from boiler import definitions, models


def invert_type_mapping(map: Dict[str, int]) -> Dict[int, str]:
    """Invert a refiner type mapping."""
    # For some reason, refiner answers provide the inverse of what you actually
    # need....
    return {v: k for k, v in map.items()}


def is_keyframe(detection: Dict[str, Any]) -> bool:
    # TODO: do we want to handle other types?
    return detection['src'] in ('ground-truth', 'truth')


def deserialize_detection(detection: Dict[str, Any]) -> models.Detection:
    box = detection['bbox']
    return models.Detection(
        frame=detection['frame'],
        box=models.Box(left=box[0], top=box[1], right=box[2], bottom=box[3],),
        keyframe=is_keyframe(detection),
    )


def deserialize_answer(answer_string: Union[str, Dict[str, Any]]) -> models.Activity:
    if isinstance(answer_string, str):
        answer = json.loads(answer_string)
    else:
        answer = answer_string

    if 'types' not in answer or 'activities' not in answer or 'detections' not in answer:
        raise BoilerError('answer did not contain all required keys')

    if len(answer['activities']) != 1:
        raise BoilerError('answer should contain exactly 1 activity')

    types = answer['types']
    activity = answer['activities'][0]
    flat_detections = sorted(answer['detections'], key=lambda d: d['id1'])

    actor_type_map: Dict[int, definitions.ActorType] = {}
    for actor_type in types:
        actor_type_string = list(actor_type['cset3'].keys())[0]
        actor_type_map[actor_type['id1']] = definitions.ActorType(actor_type_string)

    actor_map = {actor['id1']: actor for actor in activity.get('actors', [])}
    detection_map = groupby(flat_detections, lambda d: d['id1'])
    actors: List[models.Actor] = []

    for actor_id, detection_iterator in detection_map:
        answer_actor = actor_map[actor_id]
        actor_type = actor_type_map[actor_id]
        time_range = IntInterval(answer_actor['timespan'][0]['tsr0'])
        detections = [deserialize_detection(d) for d in detection_iterator if is_keyframe(d)]

        actor = models.Actor(
            begin=time_range.lower,
            end=time_range.upper,
            actor_type=actor_type.value,
            detections=detections,
        )
        actors.append(actor)

    activity_id = activity['id2']
    activity_type_map = invert_type_mapping(activity['act2'])
    activity_type = definitions.ActivityType(activity_type_map[activity_id])
    frame_range = IntInterval(activity['timespan'][0]['tsr0'])

    return models.Activity(
        activity_type=activity_type, begin=frame_range.lower, end=frame_range.upper, actors=actors,
    )



"""
Deserialize kpf to boiler_rest client types
"""
import math
from typing import Dict

import yaml

from boiler import BoilerError
from boiler.models import Activity, ActivityList, Actor, Box, Detection

GEOM_ID = 'id0'  # detection id
ACTOR_ID = 'id1'  # actor id
CSET3 = 'cset3'
GEOM = 'geom'
TYPES = 'types'
FRAME = 'ts0'  # timestamp in frames
SECONDS = 'ts1'  # timestamp in seconds
BOX = 'g0'  # geometry bounding box: [top, left, bottom, right]
KEYFRAME = 'keyframe'  # boolean is keyframe?
ACTIVITY = 'act'
ACTIVITY_TYPE = 'act2'
ACTIVITY_ID = 'id2'
TIMESPANS = 'timespan'
FRAME_TIMESPAN = 'tsr0'
STATUS = 'src_status'
ACTORS = 'actors'


def load_yaml(file):
    try:
        return yaml.load(file, Loader=yaml.CLoader)
    except AttributeError:
        return yaml.safe_load(file)


# TODO: Fix kpf deserialization and remove type: ignore


def deserialize_types(file, actor_map):
    yml = load_yaml(file)
    for type_packet in yml:
        if TYPES in type_packet:
            kpf_types = type_packet[TYPES]
            id1 = kpf_types[ACTOR_ID]
            cset3 = kpf_types[CSET3]
            cset3keys = list(cset3.keys())
            if len(cset3keys) != 1:
                raise BoilerError(f'{CSET3} should only have 1 key, found {cset3keys}')
            name = cset3keys[0]
            if id1 in actor_map:
                actor_map[id1].actor_type = name
            else:
                actor_map[id1] = Actor(actor_type=name)  # type: ignore


def deserialize_geom(file, actor_map):
    """
    python's yaml parsing is painfully slow,
    this func could take on the order of 10 seconds
    """
    yml = load_yaml(file)
    for geom_packet in yml:
        if GEOM in geom_packet:
            geom = geom_packet[GEOM]
            actor_id = geom[ACTOR_ID]
            frame = geom[FRAME]
            box = geom[BOX]
            box = [int(n) for n in box.split(' ')]
            if len(box) != 4:
                raise BoilerError('expect bounding box to have 4 values')
            keyframe = False
            if KEYFRAME in geom:
                keyframe = geom[KEYFRAME]
            box = Box(left=box[0], top=box[1], right=box[2], bottom=box[3])
            detection = Detection(frame=frame, box=box, keyframe=keyframe)
            if actor_id in actor_map:
                actor_map[actor_id].detections.append(detection)
            else:
                actor_map[actor_id] = Actor(  # type: ignore
                    clip_id=actor_id, detections=[detection]
                )

    for actor in actor_map.values():
        actor.sort_detections()


def _deserialize_frame_timespan(timespans):
    for timespan in timespans:
        if FRAME_TIMESPAN in timespan:
            return timespan[FRAME_TIMESPAN]
    raise BoilerError(f'no timespan in {timespans} with {FRAME_TIMESPAN}')


def _deserialize_actor(actor, actor_map):
    if ACTOR_ID not in actor:
        raise BoilerError(f'actor {actor} missing {ACTOR_ID}')
    if TIMESPANS not in actor:
        raise BoilerError(f'actor {actor} missing {TIMESPANS}')
    actor_id = actor[ACTOR_ID]
    timespans = actor[TIMESPANS]
    frame_timespan = _deserialize_frame_timespan(timespans)
    if actor_id in actor_map:
        actor_map[actor_id].begin = frame_timespan[0]
        actor_map[actor_id].end = frame_timespan[1]
    else:
        actor_map[actor_id] = Actor(  # type: ignore
            clip_id=actor_id, begin=frame_timespan[0], end=frame_timespan[1]
        )
    return actor_map[actor_id]


def _deserialize_activity(activity_packet, actor_map):
    """
    returns activity instance
    """
    activity = activity_packet[ACTIVITY]
    timespans = activity[TIMESPANS]
    frame_timespan = _deserialize_frame_timespan(timespans)
    actors = activity[ACTORS]
    activity_type_obj = activity[ACTIVITY_TYPE]
    activity_type_keys = list(activity_type_obj.keys())
    if len(activity_type_keys) != 1:
        raise BoilerError(f'{activity_type_obj} should only have 1 key')
    activity_type = activity_type_keys[0]
    status = activity.get(STATUS, None)
    return Activity(  # type: ignore
        clip_id=activity[ACTIVITY_ID],
        activity_type=activity_type,
        begin=frame_timespan[0],
        end=frame_timespan[1],
        status=status,
        actors=[_deserialize_actor(a, actor_map) for a in actors],
    )


def deserialize_activities(file, activity_map, actor_map):
    """
    Load activity from file given geometry and types
    returns map from activity ID to activity
    """
    yml = load_yaml(file)
    for activity_packet in yml:
        if ACTIVITY in activity_packet:
            activity = _deserialize_activity(activity_packet, actor_map)
            activity_map[activity.clip_id] = activity


def serialize_types(actor_map: Dict[int, Actor]):
    for actor_id, actor in actor_map.items():
        line = {'types': {ACTOR_ID: actor_id, CSET3: {actor.actor_type: 1.0}}}
        yield '- {}'.format(yaml.dump(line, default_flow_style=True, width=math.inf))


def serialize_geom(actor_map: Dict[int, Actor]):
    geom_id = 1
    for actor_id, a in actor_map.items():
        for d in a.detections:
            if d.keyframe:
                line = {
                    'geom': {
                        ACTOR_ID: actor_id,
                        GEOM_ID: geom_id,
                        FRAME: d.frame,
                        # TODO ts1 requires framerate
                        BOX: f'{d.box.left} {d.box.top} {d.box.right} {d.box.bottom}',
                        KEYFRAME: True,
                    }
                }
                geom_id = geom_id + 1
                yield '- {}'.format(yaml.dump(line, default_flow_style=True, width=math.inf))


def serialize_activities(activity_map: Dict[int, Activity], actor_map: Dict[int, Actor]):
    inverse_actor_map = {actor: actor_id for actor_id, actor in actor_map.items()}

    for activity_id, activity in activity_map.items():
        activity = activity_map[activity_id]

        line = {
            ACTIVITY: {
                ACTIVITY_TYPE: {activity.activity_type: 1.0,},
                ACTIVITY_ID: activity_id,
                TIMESPANS: [{FRAME_TIMESPAN: [activity.begin, activity.end]}],
                ACTORS: [
                    {
                        ACTOR_ID: inverse_actor_map[actor],
                        TIMESPANS: [{FRAME_TIMESPAN: [actor.begin, actor.end]}],
                    }
                    for actor in activity.actors
                ],
            }
        }

        if activity.status:
            line[ACTIVITY][STATUS] = activity.status

        yield '- {}'.format(yaml.dump(line, default_flow_style=True, width=math.inf))


def serialize_to_files(
    output_basename: str, activity_list: ActivityList, keyframes_only=False,
):
    actor_map = activity_list.actor_map
    activity_map = activity_list.activity_map

    with open(output_basename + '.activities.yml', 'w', encoding='utf-8') as yamlfile:
        for line in serialize_activities(activity_map, actor_map):
            yamlfile.write(line)
    with open(output_basename + '.types.yml', 'w', encoding='utf-8') as yamlfile:
        for line in serialize_types(actor_map):
            yamlfile.write(line)
    with open(output_basename + '.geom.yml', 'w', encoding='utf-8') as yamlfile:
        for line in serialize_geom(actor_map):
            yamlfile.write(line)



import datetime
from distutils.version import StrictVersion
from pathlib import Path
import platform
import sys
import traceback


import click
import requests
from requests.exceptions import RequestException
from requests_toolbelt.sessions import BaseUrlSession

__version__ = '0.0.11'


class BoilerException(Exception):
    pass


class BoilerWarning(BoilerException):
    pass


class BoilerError(BoilerException):
    pass


class BoilerSession(BaseUrlSession):

    page_size = 50

    def __init__(self, base_url, token):
        base_url = f'{base_url.rstrip("/")}/'  # tolerate input with or without trailing slash
        super(BoilerSession, self).__init__(base_url=base_url)
        self.token = token
        self.headers.update(
            {
                'User-agent': f'boiler/{__version__}',
                'Accept': 'application/json',
                'X-Stumpf-Token': self.token,
            }
        )


class KW18BaseName(click.ParamType):
    name = 'kw18_basename'

    def convert(self, value, param, ctx):
        expected_files = [
            Path(value + x) for x in ['.kw18.types', '.kw18', '.txt', '.kw18.regions']
        ]
        missing_files = [x for x in expected_files if not x.exists()]

        if missing_files:
            self.fail(f'Missing one or more expected kw18 files for {value}', param, ctx)

        return {
            'activities': {'file': open(expected_files[2], 'r'), 'model': None},
            'types': {'file': open(expected_files[0], 'r'), 'model': None},
            'regions': {'file': open(expected_files[3], 'r'), 'model': None},
            'geom': {'file': open(expected_files[1], 'r'), 'model': None},
        }


def newer_version_available():
    r = requests.get('https://pypi.org/pypi/diva-boiler/json', timeout=(5, 5))
    r.raise_for_status()
    releases = list(r.json()['releases'])
    return any(StrictVersion(r) > StrictVersion(__version__) for r in releases)


def main():
    try:
        cli()
    except Exception:
        click.echo(
            click.style(
                'The following unexpected error occurred while attempting your operation:\n',
                fg='red',
            ),
            err=True,
        )

        click.echo(traceback.format_exc(), err=True)

        click.echo(f'boiler:  v{__version__}', err=True)
        click.echo(f'python:  v{platform.python_version()}', err=True)
        click.echo(f'time:    {datetime.datetime.now(datetime.timezone.utc).isoformat()}', err=True)
        click.echo(f'os:      {platform.platform()}', err=True)
        click.echo(f'command: {" ".join(sys.argv[1:])}\n', err=True)

        click.echo(
            click.style(
                'This is a bug in boiler and should be reported here with the above output:',
                fg='yellow',
            ),
            err=True,
        )
        click.echo(
            'https://gitlab.com/diva-mturk/stumpf-diva/issues/new', err=True,
        )


@click.group()
@click.option(
    '--api-url',
    default='https://stumpf-the-younger.avidannotations.com/api/diva/',
    envvar='STUMPF_API_URL',
)
@click.option('--x-stumpf-token', envvar='X_STUMPF_TOKEN')
@click.option('--offline', is_flag=True)
@click.version_option()
@click.pass_context
def cli(ctx, api_url, x_stumpf_token, offline):
    if not offline:
        try:
            if newer_version_available():
                click.echo(
                    click.style(
                        """There is a newer version of boiler available.
You must upgrade to the latest version before continuing.
If you are using pip, then you can upgrade by running the following command:
""",
                        fg='yellow',
                    ),
                    err=True,
                )
                click.echo(click.style('pipx upgrade diva-boiler', fg='green'), err=True)
                sys.exit(1)
        except RequestException:
            click.echo(
                click.style('Failed to check for newer version of boiler:', fg='red'), err=True
            )
            raise

    session = BoilerSession(api_url, x_stumpf_token)
    ctx.obj = {'session': session}


# TODO: re-enable kpf once deserialization is fixed
from boiler.commands import activity, gunrunner, kw18, vendor, video  # noqa: F401 E402



import enum


# This is needed to get mypy to validate enum's correctly.  By default,
# enum's have value type Any, which means that it is never validated.
class StrEnum(enum.Enum):
    @property
    def value(self) -> str:
        return super(StrEnum, self).value


class ActorType(StrEnum):
    BAG = 'Bag'
    BIKE = 'Bike'
    OTHER = 'Other'
    PERSON = 'Person'
    QUEUE = 'Queue'
    RECEPTACLE = 'Receptacle'
    VEHICLE = 'Vehicle'


actor_codes = {
    ActorType.BAG: 'b',
    ActorType.BIKE: 'k',
    ActorType.OTHER: 'o',
    ActorType.PERSON: 'p',
    ActorType.QUEUE: 'q',
    ActorType.RECEPTACLE: 'o',  # https://gitlab.com/diva-mturk/stumpf-diva/issues/32#note_285508683
    ActorType.VEHICLE: 'v',
}


class ActivityType(StrEnum):
    ABANDON_PACKAGE = 'Abandon_Package'
    CARRY_IN_HANDS = 'Carry_in_Hands'
    CARRY_ON_BACK = 'Carry_on_Back'
    CASING_FACILITY = 'Casing_Facility'
    CLOSE_FACILITY_DOOR = 'Close_Facility_Door'
    CLOSE_TRUNK = 'Close_Trunk'
    CLOSE_VEHICLE_DOOR = 'Close_Vehicle_Door'
    EMBRACE_INTERACTION = 'Embrace_Interaction'
    ENTER_FACILITY = 'Enter_Facility'
    ENTER_THROUGH_STRUCTURE = 'Enter_Through_Structure'
    ENTER_VEHICLE = 'Enter_Vehicle'
    EXIT_FACILITY = 'Exit_Facility'
    EXIT_THROUGH_STRUCTURE = 'Exit_Through_Structure'
    EXIT_VEHICLE = 'Exit_Vehicle'
    HAND_INTERACTION = 'Hand_Interaction'
    HEAVY_CARRY = 'Heavy_Carry'
    JOINING_QUEUE = 'Joining_Queue'
    LAPTOP_INTERACTION = 'Laptop_Interaction'
    LOAD_VEHICLE = 'Load_Vehicle'
    OBJECT_TRANSFER = 'Object_Transfer'
    OPEN_FACILITY_DOOR = 'Open_Facility_Door'
    OPEN_TRUNK = 'Open_Trunk'
    OPEN_VEHICLE_DOOR = 'Open_Vehicle_Door'
    PEOPLE_TALKING = 'People_Talking'
    PICK_UP_OBJECT = 'Pick_Up_Object'
    PURCHASING = 'Purchasing'
    PUT_DOWN_OBJECT = 'Put_Down_Object'
    READ_DOCUMENT = 'Read_Document'
    RIDING = 'Riding'
    SIT_DOWN = 'Sit_Down'
    STAND_UP = 'Stand_Up'
    TALK_ON_PHONE = 'Talk_on_Phone'
    TEXT_ON_PHONE = 'Text_On_Phone'
    THEFT = 'Theft'
    UNLOAD_VEHICLE = 'Unload_Vehicle'
    VEHICLE_DROPSOFF_PERSON = 'Vehicle_DropsOff_Person'
    VEHICLE_MOVING = 'Vehicle_Moving'
    VEHICLE_PICKSUP_PERSON = 'Vehicle_PicksUp_Person'
    VEHICLE_REVERSING = 'Vehicle_Reversing'
    VEHICLE_STARTING = 'Vehicle_Starting'
    VEHICLE_STOPPING = 'Vehicle_Stopping'
    VEHICLE_TURNING_LEFT = 'Vehicle_Turning_Left'
    VEHICLE_TURNING_RIGHT = 'Vehicle_Turning_Right'
    VEHICLE_UTURN = 'Vehicle_UTurn'

    @property
    def value(self) -> str:
        return super(ActivityType, self).value


# activity validations are described as regular expressions
# individual actors are encoded as single characters (mapping below)
# words separated by | must be internally sorted in alphabetical order
# because candidates will be sorted before testing to avoid using regex
# lookaround.
#
# for example:
#   `bop` is a valid spec, but `pob` is not.
#   `(kpv|bo)` is valid because each individual word is sorted.
activity_spec = {
    ActivityType.ABANDON_PACKAGE: '(bp|op)',
    ActivityType.CARRY_IN_HANDS: None,
    ActivityType.CARRY_ON_BACK: None,
    ActivityType.CASING_FACILITY: None,
    ActivityType.CLOSE_FACILITY_DOOR: 'p',
    ActivityType.CLOSE_TRUNK: 'p?v',
    ActivityType.CLOSE_VEHICLE_DOOR: 'p?v',
    ActivityType.EMBRACE_INTERACTION: 'p{2,}',
    ActivityType.ENTER_FACILITY: 'p',
    ActivityType.ENTER_THROUGH_STRUCTURE: 'p',
    ActivityType.ENTER_VEHICLE: 'pv',
    ActivityType.EXIT_FACILITY: 'p',
    ActivityType.EXIT_THROUGH_STRUCTURE: 'p',
    ActivityType.EXIT_VEHICLE: 'pv',
    ActivityType.HAND_INTERACTION: 'p{2,}',
    ActivityType.HEAVY_CARRY: '(bp+|o+p+|p{2,})',
    ActivityType.JOINING_QUEUE: None,
    ActivityType.LAPTOP_INTERACTION: 'p',
    ActivityType.LOAD_VEHICLE: '(o*pv|b*pv)',
    ActivityType.OBJECT_TRANSFER: '(o*p{2}|b*p{2}|o*pv|b*pv)',
    ActivityType.OPEN_FACILITY_DOOR: 'p',
    ActivityType.OPEN_TRUNK: 'p?v',
    ActivityType.OPEN_VEHICLE_DOOR: 'pv',
    ActivityType.PEOPLE_TALKING: 'p{2,}',
    ActivityType.PICK_UP_OBJECT: '(bp|op+)',
    ActivityType.PURCHASING: 'p+',
    ActivityType.PUT_DOWN_OBJECT: '(bp|op+)',
    ActivityType.READ_DOCUMENT: 'p',
    ActivityType.RIDING: 'kp',
    ActivityType.SIT_DOWN: 'p',
    ActivityType.STAND_UP: 'p',
    ActivityType.TALK_ON_PHONE: 'p',
    ActivityType.TEXT_ON_PHONE: 'p',
    ActivityType.THEFT: '(bo*p|o+p)',
    ActivityType.UNLOAD_VEHICLE: '(o*pv|b*pv)',
    ActivityType.VEHICLE_DROPSOFF_PERSON: 'p+v',
    ActivityType.VEHICLE_MOVING: 'v',
    ActivityType.VEHICLE_PICKSUP_PERSON: 'p+v',
    ActivityType.VEHICLE_REVERSING: 'v',
    ActivityType.VEHICLE_STARTING: 'v',
    ActivityType.VEHICLE_STOPPING: 'v',
    ActivityType.VEHICLE_TURNING_LEFT: 'v',
    ActivityType.VEHICLE_TURNING_RIGHT: 'v',
    ActivityType.VEHICLE_UTURN: 'v',
}

all_activity_codes = {
    2: ActivityType.VEHICLE_MOVING,
    3: ActivityType.STAND_UP,
    4: ActivityType.SIT_DOWN,
    5: ActivityType.CARRY_IN_HANDS,
    6: ActivityType.CARRY_ON_BACK,
    7: ActivityType.TALK_ON_PHONE,
    8: ActivityType.READ_DOCUMENT,
    9: ActivityType.TEXT_ON_PHONE,
    10: ActivityType.VEHICLE_STOPPING,
    11: ActivityType.VEHICLE_STARTING,
    12: ActivityType.VEHICLE_TURNING_LEFT,
    13: ActivityType.VEHICLE_TURNING_RIGHT,
    14: ActivityType.VEHICLE_UTURN,
    15: ActivityType.OBJECT_TRANSFER,
    16: ActivityType.PICK_UP_OBJECT,
    17: ActivityType.PUT_DOWN_OBJECT,
    18: ActivityType.HEAVY_CARRY,
    19: ActivityType.HAND_INTERACTION,
    20: ActivityType.EMBRACE_INTERACTION,
    21: ActivityType.OPEN_FACILITY_DOOR,
    22: ActivityType.CLOSE_FACILITY_DOOR,
    23: ActivityType.ENTER_FACILITY,
    24: ActivityType.EXIT_FACILITY,
    25: ActivityType.OPEN_VEHICLE_DOOR,
    26: ActivityType.CLOSE_VEHICLE_DOOR,
    27: ActivityType.ENTER_VEHICLE,
    28: ActivityType.EXIT_VEHICLE,
    29: ActivityType.OPEN_TRUNK,
    30: ActivityType.CLOSE_TRUNK,
    31: ActivityType.LOAD_VEHICLE,
    32: ActivityType.UNLOAD_VEHICLE,
    33: ActivityType.VEHICLE_PICKSUP_PERSON,
    34: ActivityType.VEHICLE_DROPSOFF_PERSON,
    35: ActivityType.PEOPLE_TALKING,
    36: ActivityType.JOINING_QUEUE,
    37: ActivityType.RIDING,
    38: ActivityType.PURCHASING,
    39: ActivityType.LAPTOP_INTERACTION,
    40: ActivityType.ABANDON_PACKAGE,
    41: ActivityType.CASING_FACILITY,
    43: ActivityType.THEFT,
    44: ActivityType.VEHICLE_REVERSING,
    # Codes below this comment are made-up
    # because these activities have no documented codes
    -1000: ActivityType.ENTER_THROUGH_STRUCTURE,
    -1001: ActivityType.EXIT_THROUGH_STRUCTURE,
}


class ActivityPipelineStatuses(StrEnum):
    UNAUDITED = 'unaudited'
    AUDITED = 'audited'
    GLADIATOR_1 = 'gladiator_1'
    REFINER_1 = 'refiner_1'
    GLADIATOR_2 = 'gladiator_2'
    REFINER_2 = 'refiner_2'
    GLADIATOR_3 = 'gladiator_3'
    JANITOR = 'janitor'
    GOOD = 'good'


class AnnotationVendors(StrEnum):
    IMERIT = 'imerit'
    KITWARE = 'kitware'


class CameraLocation(StrEnum):
    ADMIN = 'admin'
    BUS = 'bus'
    SCHOOL = 'school'
    HOSPITAL = 'hospital'


class DataCollects(StrEnum):
    M1 = 'm1'
    M2 = 'm2'


class ReleaseBatches(StrEnum):
    SEQUESTERED = 'sequestered'
    MEVA = 'meva'
    TESTING = 'testing'


class CameraTypes(StrEnum):
    VISIBLE = 'eo'
    INFRARED = 'ir'


class VideoPipelineStatuses(StrEnum):
    ANNOTATION = 'annotation'
    AUDIT = 'audit'
    GUNRUNNER = 'gunrunner'
    REMEDIATION = 'remediation'
    COMPLETE = 'complete'


class Scenarios(StrEnum):
    BASKETBALL = 'basketball'
    CONFERENCE = 'conference'
    CONTROL = 'control'
    DEBATE_TEAM = 'debate team'
    DENSE_SUBWAY = 'dense subway'
    FOOTRACE = 'footrace'
    PANEL_DEBATE = 'panel debate'
    POLITICAL_DEBATE = 'political debate'
    POLITICAL_RALLY = 'political rally'
    THREAT = 'threat'
    TRAVEL = 'travel'
